# Supplementary material to the manuscript entitled
#       "Fully and Semi-Automated Shape Differentiation in NGSolve"
# by the authors
#       P. Gangl, K. Sturm, M. Neunteufel, J. Schoeberl
# submitted to the journal
#        Structural and Multidisciplinary Optimization
# Corresponding author: 
#       Peter Gangl, Institute of Applied Mathematics, 
#       Graz University of Technology, Graz, Austria
#       gangl@math.tugraz.at

nonlinear = True
order = 1

fes = VectorH1(mesh, order=1, dirichlet="left1|left3") # grad(u) ...  Jacobian

F = Id(2)
Finv = Inv(F)

force = CoefficientFunction( (0,0) )
gN = CoefficientFunction( (0,-1) )
factorFinal = 100

if nonlinear == True:
    factor = Parameter(0.1)
else:
    factor = 1
    
I = Id(mesh.dim)
def C(u):
    F1 = I + grad(u)*Finv
    return F1.trans * F1

E, nu = 1000, 0.3
mu  = E / 2 / (1+nu)
lam = E * nu / ((1+nu)*(1-2*nu))

def Pow(a, b):
    return a**b

def eps(u):
    return 1/2 * (Jacobian(u)*Finv+(Jacobian(u)*Finv).trans)

if nonlinear == True:
    def Equation_vol(u,v):
        return InnerProduct((I+grad(u)*Finv)*(lam * Trace(0.5*(C(u)-I))*I + mu*(C(u)-I)), grad(v) )*dx #St. Venant - Kirchhoff
else:
    def Equation_vol(u,v):
        return (2*InnerProduct(mu*eps(u),eps(v)) + lam*Trace(Jacobian(u)*Finv)*Trace(Jacobian(v)*Finv) )  *Det(F)*dx
    
    
def Equation_bnd(u,v):
    return (-1)*factor * InnerProduct(gN, v) *Det(F)*Norm(Inv(F).trans*specialcf.normal(2))*ds("right2")

def solvePDE():
    gfu.vec[:] = 0
    if nonlinear == True:
        Draw (gfset+gfu, mesh, "gfset_p_gfu")
        SetVisualization (deformation=True)

        nLoadSteps = 10
        for loadstep in range(nLoadSteps):
            print ("load step", loadstep)
            factor.Set (factorFinal * (loadstep+1) / nLoadSteps)
            Newton(a, gfu, maxit=20, printing = False)
            #a.AssembleLinearization(gfu.vec)
            SetVisualization (deformation=True)
    else:
        Newton(a, gfu, maxit=20, dampfactor=1, printing = False)
        a.AssembleLinearization(gfu.vec)
        
    dCostdu.Assemble()
    inva = a.mat.Inverse(fes.FreeDofs(), inverse="sparsecholesky")
    gfp.vec.data = -inva.T * dCostdu.vec
   
isComplex = False
isSurface = False
