# Supplementary material to the manuscript entitled
#       "Fully and Semi-Automated Shape Differentiation in NGSolve"
# by the authors
#       P. Gangl, K. Sturm, M. Neunteufel, J. Schoeberl
# submitted to the journal
#        Structural and Multidisciplinary Optimization
# Corresponding author: 
#       Peter Gangl, Institute of Applied Mathematics, 
#       Graz University of Technology, Graz, Austria
#       gangl@math.tugraz.at

order = 1
fes = H1(mesh, order=order, complex=True)

# Wavenumber & source
omega = 100
pulse = 1e3*exp(-(3**2)*((X[0]-0.2)*(X[0]-0.2) + (X[1]-0.5)*(X[1]-0.5)))
Draw(pulse, mesh, 'pulse')

# for shape differentiation
F = Id(2)
FinvT = Inv(F).trans
myN = specialcf.normal(2)
xi_tang = Det(F)*sqrt((FinvT*myN)*(FinvT*myN))


def Equation_lhs_vol(u,v):
    return ( InnerProduct(FinvT*grad(u), FinvT*grad(v) ) - omega**2*InnerProduct(u,v) )*Det(F)*dx
def Equation_lhs_bnd(u,v):
    return (-1)*omega*1j*InnerProduct(u,v) *xi_tang* ds("ne|nw|se|sw")
def Equation_rhs(v):
    return InnerProduct(pulse, v) * Det(F)*dx

def solvePDE():
    gfu.vec[:] = 0        
    ca.Update()
    a.Assemble()
    f.Assemble()
    solvers.BVP(bf=a, lf=f, gf=gfu, pre=ca, print=False)
   
    dCostdu.Assemble()
    a.AssembleLinearization(gfu.vec)
    gfp.vec.data = -a.mat.Inverse(fes.FreeDofs()).T * dCostdu.vec
    
    #input("pause")
    
def RealPart (igls):
    l = [igl.coef.real*igl.symbol for igl in igls]
    return sum(l[1:],l[0])

isComplex = True
isSurface = False
