# Supplementary material to the manuscript entitled
#       "Fully and Semi-Automated Shape Differentiation in NGSolve"
# by the authors
#       P. Gangl, K. Sturm, M. Neunteufel, J. Schoeberl
# submitted to the journal
#        Structural and Multidisciplinary Optimization
# Corresponding author: 
#       Peter Gangl, Institute of Applied Mathematics, 
#       Graz University of Technology, Graz, Austria
#       gangl@math.tugraz.at

from ngsolve import * 
from netgen.csg import *
from ngsolve.internal import * 

def generateGeometry(myZlength, bCoarse):
    zlength = myZlength
    front = Plane (Pnt(0,0,zlength/2), Vec(0,0,1) )
    back = Plane (Pnt(0,0,-zlength/2), Vec(0,0,-1) )
    r1 = 4
    r2 = 8
    r3 = 19
    r4 = 19.5000356070342
    r5 = 19.6666670000000
    r6 = 19.8333330000000
    r7 = 20
    r8 = 23.5
    r9 = 27.5
    r10 = 33
    cubeleft = Plane( Pnt(0,0,0), Vec(-1,0,0) )
    cubebot = Plane( Pnt(0,0,0), Vec(0,-1,0) )
    cyl1 = Cylinder ( Pnt(0, 0, -zlength/2), Pnt(0, 0, zlength/2), r1) 
    cyl2 = Cylinder ( Pnt(0, 0, -zlength/2), Pnt(0, 0, zlength/2), r2) 
    cyl3 = Cylinder ( Pnt(0, 0, -zlength/2), Pnt(0, 0, zlength/2), r3) 
    cyl4 = Cylinder ( Pnt(0, 0, -zlength/2), Pnt(0, 0, zlength/2), r4) 
    cyl5 = Cylinder ( Pnt(0, 0, -zlength/2), Pnt(0, 0, zlength/2), r5) 
    cyl6 = Cylinder ( Pnt(0, 0, -zlength/2), Pnt(0, 0, zlength/2), r6)
    cyl7 = Cylinder ( Pnt(0, 0, -zlength/2), Pnt(0, 0, zlength/2), r7)
    cyl8 = Cylinder ( Pnt(0, 0, -zlength/2), Pnt(0, 0, zlength/2), r8)
    cyl9 = Cylinder ( Pnt(0, 0, -zlength/2), Pnt(0, 0, zlength/2), r9)
    cyl10 = Cylinder ( Pnt(0, 0, -zlength/2), Pnt(0, 0, zlength/2), r10)

    cyl12 = (cyl2-cyl1)*front*back
    cyl23 = (cyl3-cyl2)*front*back
    cyl34 = (cyl4-cyl3)*front*back
    cyl45 = (cyl5-cyl4)*front*back
    cyl56 = (cyl6-cyl5)*front*back
    cyl67 = (cyl7-cyl6)*front*back
    cyl78 = (cyl8-cyl7)*front*back
    cyl89 = (cyl9-cyl8)*front*back
    cyl910 = (cyl10-cyl9)*front*back

    front2 = Plane (Pnt(0,0,zlength/2+1), Vec(0,0,1) )
    back2 = Plane (Pnt(0,0,-zlength/2-1), Vec(0,0,-1) )
    cyltotal = ((cyl10 - cyl1)*front2*back2).bc("outer") #we want no homo Dirichlet conditions on front/back
    cyltotaltotal = ((cyl5 - cyl1)*front*back* cubeleft * cubebot ).bc("outerVEC")
    cyltotaltotal = ((cyl6 - cyl5)*front*back* cubeleft * cubebot ).bc("outerVEC")
    cyltotaltotal = ((cyl10 - cyl6)*front*back* cubeleft * cubebot ).bc("outerVEC")

    dom11 = cyl12 * cubeleft * cubebot 
    dom6 = cyl34 * cubeleft * cubebot 
    dom12 = cyl45 * cubeleft * cubebot
    dom13 = cyl56 * cubeleft * cubebot 
    dom14 = cyl67 * cubeleft * cubebot
    dom15 = cyl78 * cubeleft * cubebot 
    dom16 = cyl89 * cubeleft * cubebot
    dom17 = cyl910 * cubeleft * cubebot 
    #dom5 = cyl23 * cubeleft * cubebot

    p1 = Pnt(0.500000000065054, 13.645849950289424, 0) #p[57]
    p2 = Pnt(10.443569941452147, 12.774265023038632, 0) #p[72]
    p12 = Pnt(1.64805029716032, 16.4174885478233, 0) #p[59]
    p21 = Pnt(9.29551964435688, 10.0026264255048, 0) #p[70]
    p3 = Pnt(0.500000000027635, 14.1458499502447, 0) #p[53]
    p4 = Pnt(0.707106781233923, 14.1458499502467, 0) #p[58]
    p5 = Pnt(9.64907303494498, 10.3561798160400, 0) #[74]
    p6 = Pnt(9.50262642552575, 10.5026264254621, 0) #p[71]
    p7 = Pnt(2.41377358864114, 18.8460525591577, 0)  #%p[51]
    p8 = Pnt(0.500000000080680, 18.9934199131879, 0) #%p[41]
    p9 = Pnt(11.6193758904814, 15.0329672358937, 0)# %p[76]
    p10 = Pnt(13.0768226279808, 13.7839294091508, 0) #%p[86]

    q1 = Pnt(10.002626425543037, 9.295519644264491, 0); 
    q2 = Pnt(16.417488547861552, 1.648050297067939, 0);
    q12 = Pnt(12.774265023076905, 10.443569941359774, 0);
    q21 = Pnt(13.645849950327715, 0.499999999972688, 0);
    q3 = Pnt(10.356179816078228,   9.649073034852600, 0);
    q4 = Pnt(10.502626425500326,   9.502626425433329, 0);
    q5 = Pnt(14.145849950282992,   0.499999999935281, 0);
    q6 = Pnt(14.145849950285022,   0.707106781141541, 0);
    q7 = Pnt(15.032967235955637,  11.619375890401361, 0);
    q8 = Pnt(13.783929409189092,  13.076822627888449, 0);
    q9 = Pnt(18.846052559170502,   2.413773588540749, 0);
    q10 = Pnt(18.993419913226123,   0.499999999988301, 0);

    n_1_12 = Vec(-2.771638597533876,  1.148050297095266, 0)
    n_12_2 = Vec( 3.643223524784668,  8.795519644291826, 0)
    n_2_21 = Vec( 2.771638597533832, -1.148050297095267, 0)
    n_21_1 = Vec(-3.643223524784624, -8.795519644291826, 0)
    n_4_3  = Vec( 0.000000000001998, -0.207106781206288, 0) 
    n_3_8  = Vec(-4.847569962943199,  0.000000000053045, 0)
    n_7_12 = Vec( 2.428564011334402, -0.765723291480820, 0)
    n_12_4 = Vec( 2.271638597576601, -0.940943515926397, 0)
    n_2_9  = Vec(-2.258702212855068,  1.175805949029254, 0)
    n_6_2  = Vec(-2.271638597576532,  0.940943515926396, 0)
    n_5_6  = Vec(-0.146446609422101, -0.146446609419229, 0)
    n_10_5 = Vec( 3.427749593110800, -3.427749593035820, 0)

    n_q_1_12 = Vec(-1.148050297095283,   2.771638597533869, 0)
    n_q_12_2 = Vec(8.795519644291835,   3.643223524784647, 0)
    n_q_2_21 = Vec(1.148050297095251,  -2.771638597533837, 0)
    n_q_21_1 = Vec(-8.795519644291803,  -3.643223524784679, 0)
    n_q_4_3  = Vec(-0.146446609419272,  -0.146446609422098, 0) 
    n_q_3_8  = Vec(-3.427749593035848,   3.427749593110864, 0)
    n_q_7_12 = Vec(1.175805949041587,  -2.258702212878731, 0)
    n_q_12_4 = Vec(0.940943515926445,  -2.271638597576580, 0)
    n_q_2_9  = Vec(-0.765723291472810,   2.428564011308950, 0)
    n_q_6_2  = Vec(-0.940943515926398,   2.271638597576530, 0)
    n_q_5_6  = Vec(-0.207106781206260,   0.000000000002030, 0)
    n_q_10_5 = Vec(0.000000000053020,  -4.847569962943131, 0)

    mag1_dir = Vec(0.382683432365087,   0.923879532511288, 0)
    mag2_dir = Vec(-0.923879532511286,  -0.382683432365093, 0)

    #mag1
    left  = Plane (p1, n_1_12 )
    top   = Plane (p12, n_12_2 )
    right = Plane (p2, n_2_21 )
    bot   = Plane (p21, n_21_1 )
    dom1 = left * right * front * back * bot * top

    #mag2
    left  = Plane (q1, n_q_1_12 )
    top   = Plane (q12, n_q_12_2 )
    right = Plane (q2, n_q_2_21 )
    bot   = Plane (q21, n_q_21_1 )
    dom2 = left * right * front * back * bot * top

    #shoe7
    left = Plane ( p3, n_3_8 )
    bot = Plane ( p4, n_4_3 )
    rightbot = Plane ( p12, n_12_4 )
    righttop = Plane ( p12, n_7_12 )
    dom7 = left * bot * rightbot * righttop * (cyl3 * cubeleft*cubebot) * front * back

    #shoe8
    lefttop = Plane ( p2, n_2_9 )
    leftbot = Plane ( p6, n_6_2 )
    bot     = Plane ( p5, n_5_6 )
    right   = Plane ( p5, n_10_5)
    dom8    = lefttop * leftbot * bot * right * (cyl3 * cubeleft*cubebot) * front * back

    #shoe9
    left = Plane ( q3, n_q_3_8 )
    bot = Plane ( q4, n_q_4_3 )
    rightbot = Plane ( q12, n_q_12_4 )
    righttop = Plane ( q12, n_q_7_12 )
    dom9 = left * bot * rightbot * righttop * (cyl3 * cubeleft*cubebot) * front * back

    #shoe10
    lefttop = Plane ( q2, n_q_2_9 )
    leftbot = Plane ( q6, n_q_6_2 )
    bot     = Plane ( q5, n_q_5_6 )
    right   = Plane ( q5, n_q_10_5)
    dom10    = lefttop * leftbot * bot * right * (cyl3 * cubeleft*cubebot) * front * back

    #des3
    left = Plane ( p12, -n_7_12 )
    bot = Plane (p12, -n_12_2 )
    bot2 = Plane (p4, -n_12_2 )
    right = Plane ( p2, -n_2_9 )
    dom3 = left * bot * right * (cyl3 * cubeleft*cubebot) * front * back

    #des4
    left = Plane ( q12, -n_q_7_12 )
    bot = Plane (q12, -n_q_12_2 )
    bot2 = Plane (q4, -n_q_12_2 )
    right = Plane ( q2, -n_q_2_9 )
    dom4 = left * bot * right * (cyl3 * cubeleft*cubebot) * front * back

    #dom5  (== iron core rotor (without design areas and iron ring))
    left     = Plane ( p3, -n_3_8 )
    botleft  = Plane ( p4, -n_4_3 )
    leftbot  = Plane ( p1, -n_1_12 )
    bot      = Plane ( p21,-n_21_1 )
    rightbot = Plane ( p2, -n_2_21 )
    botright = Plane ( p5, -n_5_6 )
    right    = Plane ( p5, -n_10_5 )
    left_q     = Plane ( q3, -n_q_3_8 )
    botleft_q  = Plane ( q4, -n_q_4_3 )
    leftbot_q  = Plane ( q1, -n_q_1_12 )
    bot_q      = Plane ( q21,-n_q_21_1 )
    rightbot_q = Plane ( q2, -n_q_2_21 )
    botright_q = Plane ( q5, -n_q_5_6 )
    right_q    = Plane ( q5, -n_q_10_5 )
    dom5 = (left+botleft) * (leftbot + bot + rightbot) * (botright+right) * (left_q+botleft_q) * (leftbot_q + bot_q + rightbot_q) * (botright_q+right_q) * cyl23 * cubeleft*cubebot 

    geo = CSGeometry()
    geo.Add(dom11.mat("air").transp())
    geo.Add(dom6.mat("iron"))
    geo.Add(dom12.mat("air").transp())
    geo.Add(dom13.mat("airgap").transp())
    geo.Add(dom14.mat("air").transp())
    geo.Add(dom15.mat("air").transp())
    geo.Add(dom16.mat("iron"))
    geo.Add(dom17.mat("air").transp())

    geo.Add(dom1.mat("magnet1"))
    geo.Add(dom2.mat("magnet2"))
    if bCoarse:
        geo.Add(dom3.mat("iron").maxh(0.6))
        geo.Add(dom4.mat("iron").maxh(0.6))
        #geo.Add(dom3.mat("iron").maxh(2))
        #geo.Add(dom4.mat("iron").maxh(2))
    else:
        geo.Add(dom3.mat("iron").maxh(0.4))
        geo.Add(dom4.mat("iron").maxh(0.4))
    geo.Add(dom5.mat("iron"))

    geo.Add(dom7.mat("air").transp())
    geo.Add(dom8.mat("air").transp())
    geo.Add(dom9.mat("air").transp())
    geo.Add(dom10.mat("air").transp())

    geo.PeriodicSurfaces(cubeleft, cubebot, trafo=Trafo(Pnt(0,0,0),Vec(0,0,-1), 3.14159265358979323846/2))
    
    return geo

bCoarse = True

myzlength = 5 #40
geo = generateGeometry(myzlength, bCoarse)

mesh = Mesh(geo.GenerateMesh(maxh=1.5))
mesh.Curve(3)


ngsolve.internal.visoptions.clipsolution="vec"
ngsolve.internal.viewoptions.clipping.enable=1
ngsolve.internal.viewoptions.clipping.ny=0
ngsolve.internal.viewoptions.clipping.nz=-1
ngsolve.internal.viewoptions.clipping.dist=0
