# Supplementary material to the manuscript entitled
#       "Fully and Semi-Automated Shape Differentiation in NGSolve"
# by the authors
#       P. Gangl, K. Sturm, M. Neunteufel, J. Schoeberl
# submitted to the journal
#        Structural and Multidisciplinary Optimization
# Corresponding author: 
#       Peter Gangl, Institute of Applied Mathematics, 
#       Graz University of Technology, Graz, Austria
#       gangl@math.tugraz.at

import sys

import netgen.gui
import time

from ngsolve import *
from netgen.geom2d import unit_square
from netgen.geom2d import SplineGeometry

from ngsolve.internal import visoptions
from ngsolve.solvers import *
from math import pi, cos, sin

ngsglobals.msg_level = 1
visoptions.vecfunction='gfset'
SetVisualization (deformation=True)
X = CoefficientFunction((x,y,z))

######################################################
######################################################
exampleNumber = 5#2
######################################################
#1  int f (use other main file)
#2  Poisson
#3  Nonlinear elasticity 
#4  Helmholtz 
#5  Electric motor 3D
#6  Surface PDE
########################################################
if exampleNumber == 1:
    print("Please use other main file.")
    sys.exit(0)
elif exampleNumber == 2:
    exec(open("2a_geo.py").read())
    exec(open("2b_pde.py").read())
    exec(open("2c_cost.py").read())
    exec(open("2d_shapeopt.py").read())
elif exampleNumber == 3:
    exec(open("3a_geo.py").read())
    exec(open("3b_pde.py").read())
    exec(open("3c_cost.py").read())
    exec(open("3d_shapeopt.py").read())
elif exampleNumber == 4:
    exec(open("4a_geo.py").read())
    exec(open("4b_pde.py").read())
    exec(open("4c_cost.py").read())
    exec(open("4d_shapeopt.py").read())
elif exampleNumber == 5:
    exec(open("5a_geo.py").read())
    exec(open("5b_pde.py").read())
    exec(open("5c_cost.py").read())
    exec(open("5d_shapeopt.py").read())
elif exampleNumber == 6:
    exec(open("6a_geo.py").read())
    exec(open("6b_pde.py").read())
    exec(open("6c_cost.py").read())
    exec(open("6d_shapeopt.py").read())
elif exampleNumber == 7:
    print("Please use other main file.")
    sys.exit(0)
else:
    print("Invalid example number. Insert integer between 1 and 6")
    sys.exit()

############### Checking input... ############
def checkInput(stream=None):
    if stream is None:
        stream = sys.stdout
        
    stream.write("isComplex == " + str(isComplex) + "\n" )
    stream.write("isSurface == " + str(isSurface) + "\n" )
    #optimization parameters
    stream.write("newton == " + str(newton) +"\n")
    if newton == False:
        stream.write("lineSearch == " + str( lineSearch)  +"\n")
        if lineSearch == True:
            stream.write("armijo_param == " + str( armijo_param)  +"\n")
            stream.write("scaleIncreaseFactor == " + str( scaleIncreaseFactor)  +"\n")
        stream.write("scale_init == " + str( scale_init)  +"\n")
        stream.write("myAlpha == " + str(myAlpha)  +"\n")
    else:   #newton == True
        stream.write("regExt == " + str( regExt)  +"\n")
        stream.write("delta_regHess == " + str( delta_regHess)  +"\n")
        stream.write("nDampedSteps == " + str( nDampedSteps)  +"\n")
        
    stream.write("CR == " + str( CR)  +"\n")
    if CR == True:
        stream.write("gamma_CR == " + str( gamma_CR)  +"\n")
    
    stream.write("moveMesh == " + str(moveMesh)  +"\n")
    if moveMesh == True:
        stream.write("nMeshOpti == " + str(nMeshOpti)  +"\n")
        

    stream.write("iter_max == " + str( iter_max)  +"\n")
    stream.write("tol_scale == " + str( tol_scale)  +"\n")
    stream.write("tol_gfX == " + str( tol_gfX)  +"\n")
    stream.write("tol_J == " + str( tol_J)  +"\n")
    
    stream.write("\n ndofs == " + str(fes.ndof)  +"\n")
    stream.write("nVerts == " + str(mesh.nv)  +"\n")
    stream.write("nElems == " + str(mesh.ne)  +"\n")
    stream.write("nFaces == " + str(mesh.nface)  +"\n")
    stream.write("polyDegree == " + str(order)  +"\n")
    

#write parameters to file (create folder named "output" manually first)
outputFolder = "output"
fileName1 = outputFolder + "/params.txt"
myFile1 = open(fileName1,"w+")
checkInput(myFile1)
myFile1.close()

checkInput()        #writes parameters to console
###########################
def getExtension(gfX_bnd, freedofs, gfX_ext):
    u,v = VEC.TnT()
    aX_ext = BilinearForm(VEC)
    aX_ext += InnerProduct(grad(u)+grad(u).trans, grad(v))*dx# + InnerProduct(u,v)*dx

    gfX_ext.Set(gfX_bnd, BND)

    aX_ext.Assemble()

    r = gfX_bnd.vec.CreateVector()
    r.data = (-1) * aX_ext.mat * gfX_bnd.vec

    gfX_ext.vec.data += aX_ext.mat.Inverse(freedofs=freedofs) * r
###########################################
def moveNGmesh(displ, mesh):
    for p in mesh.ngmesh.Points():
        mip = mesh(p[0],p[1])
        v = displ(mip)
        p[0] += v[0]
        p[1] += v[1]
    mesh.ngmesh.Update()    
###########################################

VEC = VectorH1(mesh, order=order, dirichlet = dirichletOpti)

freeDofsCombined=BitArray(VEC.ndof + 2*fes.ndof)
for i in range(VEC.ndof):
    freeDofsCombined[i] = not VEC.FreeDofs()[i]
for i in range(fes.ndof):
    freeDofsCombined[VEC.ndof+i] = fes.FreeDofs()[i]
    freeDofsCombined[VEC.ndof+fes.ndof+i] = fes.FreeDofs()[i]

if newton == True:
    X3 = FESpace([VEC, fes, fes])

    # define trial and test functions for newton system
    PHI, w, q = X3.TrialFunction()
    PSI, uTest, pTest = X3.TestFunction()

# for first order method
PHIVEC = VEC.TrialFunction()
PSIVEC = VEC.TestFunction()

# define trial and test functions for state and adjoint problems
u, v = fes.TnT()
pTest1 = fes.TestFunction()

gfset = GridFunction(VEC)
gfsetTemp = GridFunction(VEC)
if mesh.dim==2:
    gfset.Set((0,0))
elif mesh.dim == 3:
    gfset.Set((0,0,0))

gfX = GridFunction(VEC)
gfX_bnd = GridFunction(VEC)
gfu = GridFunction(fes)
gfp = GridFunction(fes)

a = BilinearForm(fes)
if isComplex == False:
    a += Equation_vol(u, v)
    G_vol = Cost_vol(gfu) + Equation_vol(gfu,gfp)
    a += Equation_bnd(u,v)
    G_bnd=Cost_bnd(gfu)+Equation_bnd(gfu,gfp)
    ca = Preconditioner(a,"direct", inverse="sparsecholesky")
else:
    a += Equation_lhs_vol(u, v)
    a += Equation_lhs_bnd(u,v)
    f = LinearForm(fes)
    f += Equation_rhs(v)
    G_vol = Cost_vol(gfu) + Equation_lhs_vol(gfu,gfp) + Equation_rhs(gfp)
    G_bnd = Cost_bnd(gfu) + Equation_lhs_bnd(gfu, gfp)
    ca = Preconditioner(a,"direct") #<- sparse direct solver


Draw(gfX,mesh,"gfX")
Draw (gfp, mesh, "adjoint")
Draw (gfu, mesh, "state")
if exampleNumber == 5:
    Draw(curl(gfu), mesh, "curl(gfu)")
Draw(gfset,mesh,"gfset")

#define right hand side for adjoint equation
dCostdu = LinearForm(fes)
dCostdu += Cost_vol(gfu).Diff(gfu,pTest1)
if costbnd == True:
    dCostdu += Cost_bnd(gfu).Diff(gfu,pTest1)

if newton == True:
    gfCombined = GridFunction(X3)
    
    shapeHessLag = BilinearForm(X3)
    
    shapeHessLag+= ( G_vol.Diff(F,grad(PHI)) +  G_vol.Diff(X, PHI) ).Diff(F,grad(PSI))
    shapeHessLag+= ( G_vol.Diff(F,grad(PHI)) +  G_vol.Diff(X, PHI) ).Diff(X, PSI)
    
    shapeHessLag+= ( G_bnd.Diff(F,grad(PHI).Trace()) +  G_bnd.Diff(X, PHI) ).Diff(F,grad(PSI).Trace())
    shapeHessLag+= ( G_bnd.Diff(F,grad(PHI).Trace()) +  G_bnd.Diff(X, PHI) ).Diff(X, PSI)
    
    if regExt == True:
        shapeHessLag += delta_regHess * InnerProduct(PHI, specialcf.tangential(mesh.dim)) * InnerProduct(PSI, specialcf.tangential(mesh.dim))*ds
    else:
        shapeHessLag += delta_regHess * InnerProduct(grad(PHI), grad(PSI)) * dx
        shapeHessLag += delta_regHess * InnerProduct(PHI, PSI) * dx
        
    if CR == True:
        shapeHessLag += gamma_CR * (grad(PHI)[0,0] - grad(PHI)[1,1])*(grad(PSI)[0,0] - grad(PSI)[1,1]) *dx
        shapeHessLag += gamma_CR * (grad(PHI)[1,0] + grad(PHI)[0,1])*(grad(PSI)[1,0] + grad(PSI)[0,1]) *dx
    
    #block 1,2
    shapeHessLag += (G_vol.Diff(F, grad(PSI)) + G_vol.Diff(X, PSI)).Diff(gfu, w)
    shapeHessLag += (G_bnd.Diff(F, grad(PSI).Trace()) + G_bnd.Diff(X, PSI)).Diff(gfu, w)
        
    #block 1,3
    shapeHessLag += (G_vol.Diff(F, grad(PSI)) + G_vol.Diff(X, PSI)).Diff(gfp, q)
    shapeHessLag += (G_bnd.Diff(F, grad(PSI).Trace()) + G_bnd.Diff(X, PSI)).Diff(gfp, q)
    
    #block 2,1 
    shapeHessLag += (G_vol.Diff(gfu, uTest)).Diff(F, grad(PHI)) + (G_vol.Diff(gfu, uTest)).Diff(X, PHI)
    shapeHessLag += (G_bnd.Diff(gfu, uTest)).Diff(F, grad(PHI).Trace()) + (G_bnd.Diff(gfu, uTest)).Diff(X, PHI)
    
    #block 2,2
    shapeHessLag += (G_vol.Diff(gfu, uTest)).Diff(gfu, w)
    shapeHessLag += (G_bnd.Diff(gfu, uTest)).Diff(gfu, w)
    
    #block 2,3
    shapeHessLag += (G_vol.Diff(gfu, uTest)).Diff(gfp, q)
    shapeHessLag += (G_bnd.Diff(gfu, uTest)).Diff(gfp, q)
    
    #block 3,1
    shapeHessLag += (G_vol.Diff(gfp,pTest)).Diff(F, grad(PHI)) + (G_vol.Diff(gfp,pTest)).Diff(X, PHI)
    shapeHessLag += (G_bnd.Diff(gfp,pTest)).Diff(F, grad(PHI).Trace()) + (G_bnd.Diff(gfp,pTest)).Diff(X, PHI) 
    
    #block 3,2
    shapeHessLag += (G_vol.Diff(gfp,pTest) ).Diff(gfu, w)    
    shapeHessLag += (G_bnd.Diff(gfp,pTest) ).Diff(gfu, w)
    
    cHess = Preconditioner(shapeHessLag, "direct")
    #####
    shapeGradLag = LinearForm(X3)
    
    #line 1
    shapeGradLag += G_vol.Diff(F, grad(PSI)) + G_vol.Diff(X, PSI)
    shapeGradLag += G_bnd.Diff(F, grad(PSI).Trace()) + G_bnd.Diff(X, PSI)
    
    #line 2
    shapeGradLag += G_vol.Diff(gfu, uTest)    
    shapeGradLag += G_bnd.Diff(gfu, uTest)
    
    #line 3
    shapeGradLag += G_vol.Diff(gfp, pTest)    
    shapeGradLag += G_bnd.Diff(gfp, pTest)
    
else:   #i.e. newton == False
    aX = BilinearForm(VEC)    
    if isSurface == False:
        aX += myAlpha * InnerProduct(grad(PHIVEC), grad(PSIVEC))*dx + InnerProduct(PHIVEC,PSIVEC)*dx
        if CR == True:
            aX += gamma_CR * (grad(PHIVEC)[0,0] - grad(PHIVEC)[1,1])*(grad(PSIVEC)[0,0] - grad(PSIVEC)[1,1]) *dx
            aX += gamma_CR * (grad(PHIVEC)[1,0] + grad(PHIVEC)[0,1])*(grad(PSIVEC)[1,0] + grad(PSIVEC)[0,1]) *dx
    else:
        aX += ( InnerProduct(grad(PHIVEC).Trace(), grad(PSIVEC).Trace()) + PHIVEC.Trace()*PSIVEC.Trace())*ds
    
    
    fX = LinearForm(VEC)    
    if isComplex==False:
        #fX += G_vol.DiffShape(PSIVEC)
        #fX += G_bnd.DiffShape(PSIVEC)
        
        fX += G_vol.Diff(F, grad(PSIVEC))
        fX += G_vol.Diff(X, PSIVEC)
        
        fX += G_bnd.Diff(F, grad(PSIVEC).Trace())
        fX += G_bnd.Diff(X, PSIVEC)
    else:
        #fX += RealPart( G_vol.DiffShape(PSIVEC) )
        #fX += RealPart( G_bnd.DiffShape(PSIVEC) )    
    
        fX += RealPart( G_vol.Diff(F, grad(PSIVEC)) )
        fX += RealPart( G_vol.Diff(X, PSIVEC) )
        
        fX += RealPart( G_bnd.Diff(F, grad(PSIVEC).Trace()) )
        fX += RealPart( G_bnd.Diff(X, PSIVEC) )


if newton == False:
    aX.Assemble()
    invaX = aX.mat.Inverse(VEC.FreeDofs(), inverse="sparsecholesky")
    
##################### prepare output ###############
fileName = outputFolder + "/Jvals.m"
myFile = open(fileName,"w+")
myFile.write("Jvals_" + str(exampleNumber) + " = [")
myFile.close()
##################### prepare output ###############
    
input("press enter to start optimization")

t_optimization = time.time()
if newton == True or lineSearch == False:
    if newton == True:
        scale = 1
    else:
        scale = scale_init
    with TaskManager():
        for k in range(iter_max):
            if moveMesh == False:
                mesh.SetDeformation(gfset)
            solvePDE()
            currentJ = Integrate (Cost_vol(gfu)+Cost_bnd(gfu), mesh)
            print("it", k, 'cost', currentJ)

            if newton == True:
                shapeGradLag.Assemble()
                shapeHessLag.Assemble()
                if regExt:
                    gfCombined.vec.data = shapeHessLag.mat.Inverse(freedofs=freeDofsCombined, inverse="umfpack") * shapeGradLag.vec
                    gfX_bnd.vec.data = gfCombined.components[0].vec
                    getExtension(gfX_bnd, VEC.FreeDofs(), gfX)
                else:
                    gfCombined.vec.data = shapeHessLag.mat.Inverse(freedofs=X3.FreeDofs(), inverse="umfpack") * shapeGradLag.vec
                    gfX.vec.data = gfCombined.components[0].vec
            else:
                aX.Assemble()
                fX.Assemble()
                invaX.Update()
                gfX.vec.data = invaX * fX.vec
                
            #damped newton for regExt            
            if newton == True and nDampedSteps > 0 and k < nDampedSteps:
                scale = (k+1) / nDampedSteps
            
            if moveMesh == True:
                moveNGmesh(-scale * gfX, mesh)
                VEC.Update()
                fes.Update()
            else:
                gfset.vec.data -= scale * gfX.vec
            
            currentNormGFX = Norm(gfX.vec)
            print("------------------------------------------||gfX||", currentNormGFX )
            
            myFile = open(fileName,"a+")
            myFile.write(str(k) + " " + str(currentJ) + " " + str(currentNormGFX) + ";" )  
            myFile.close()
            
            Redraw(blocking=True)
            
            if scale < tol_scale or currentNormGFX < tol_gfX or currentJ.real < tol_J:
                break
        
        myFile = open(fileName,"a+")
        myFile.write("]; \n")
        myFile.close()

else:   #newton == False and lineSeach == True
    with TaskManager():        
        isConverged = False
        if moveMesh == False:
            mesh.SetDeformation(gfset)
        solvePDE()
        if exampleNumber == 5:        #only for motor
            writeBrValues(gfset, 0)
        Jnew = Integrate (Cost_vol(gfu)+Cost_bnd(gfu), mesh)
        if moveMesh == False:
            mesh.UnsetDeformation()
        Jold = Jnew
        print("it init", 'cost', Jnew )
        
        scale = scale_init 
        for k in range(iter_max):
            if moveMesh == False:
                mesh.SetDeformation(gfset)
            aX.Assemble()
            fX.Assemble()
            invaX = aX.mat.Inverse(VEC.FreeDofs(), inverse="sparsecholesky")
            gfX.vec.data = invaX * fX.vec
            
            currentNormGFX = Norm(gfX.vec)
            myFile = open(fileName,"a+")
            myFile.write(str(k) + " " + str(Jnew) + " " + str(currentNormGFX) + ";" )  
            myFile.close()
            
            while True:
                if scale < tol_scale or currentNormGFX < tol_gfX or Jnew.real < tol_J:
                    isConverged = True
                    break
            
                if moveMesh == False:
                    gfsetTemp.vec.data = gfset.vec - scale * gfX.vec
                    mesh.SetDeformation(gfsetTemp)
                else:
                    moveNGmesh(-scale * gfX, mesh)
                    VEC.Update()
                    fes.Update()
                    
                solvePDE()
              
                Jnew = Integrate (Cost_vol(gfu)+Cost_bnd(gfu), mesh)
                if moveMesh == False:
                    mesh.UnsetDeformation()
                    
                print("Jnew = ", Jnew.real, "condition = ", Jold.real - armijo_param*scale * currentNormGFX**2)
                if Jnew.real < Jold.real - armijo_param*scale * currentNormGFX**2 :
                    Jold = Jnew
                    print("----------it", k, 'scale', scale, 'cost', Jnew )
                    if moveMesh == False:
                        gfset.vec.data = gfsetTemp.vec
                    scale *= scaleIncreaseFactor
                    
                    if exampleNumber == 5:  #only for motor
                        writeBrValues(gfset, k+1)
                    break
                else:
                    if moveMesh == True:
                        moveNGmesh(scale * gfX, mesh)
                        VEC.Update()
                        fes.Update()
                        solvePDE()      #instead of interpolating grid functions to new mesh
                    scale = scale / 2
                
            print("------------------------------------------||gfX||", currentNormGFX )
            print("isConverged", isConverged)
            Redraw(blocking=True)
            if isConverged== True:
                print("converged with J = ", Jnew.real, ", ||gfX||=", currentNormGFX, ", scale = ", scale)
                print("Thank you for using NGShapeOpt")
                break
        myFile = open(fileName,"a+")
        myFile.write("]; \n")
        myFile.close()

t_optimization = time.time() - t_optimization

myFile = open(fileName,"a+")
myFile.write("yyaxis left; plot(Jvals_" + str(exampleNumber) + "(:,1), Jvals_" + str(exampleNumber) +"(:,2)); ylabel('J') \n")
myFile.write("yyaxis right; plot( Jvals_" + str(exampleNumber) +"(:,1),  Jvals_" + str(exampleNumber) +"(:,3),'--'); ylabel('$\|\ nabla J\|$', 'Interpreter', 'latex') \n")
myFile.write("title('History of ...');  \n xlabel('Iterations') \n")
myFile.write("%  optimization time: " + str(t_optimization) + " seconds")
myFile.close()
breakpoint = input("Press any key to quit")
